#
# (c) 1998 Michel Beaudouin-Lafon, mbl@daimi.au.dk
#
# small example of instrumental interaction to search strings in a text
#

source sm.tcl

wm title . "Editor"

# create a text window with a scrollbar
#text .txt -width 80 -height 40 -yscrollcommand {.vsb set}
#scrollbar .vsb -command {.txt yview}
text .txt -width 80 -height 40 -yscrollcommand {YScroll}
scrollbar .vsb -command {YView}
proc YScroll {args} {}
proc YView {args} {}

# display dialog box to load a text file
if [file readable "search.txt"] {
	set filename "search.txt"
} else {
	set filename [tk_getOpenFile -title "Load File"]
}
if {$filename != ""} {
	set fd [open $filename r]
	.txt insert end [read $fd]
	close $fd
}

# create the search dialog box
frame .search
label .search.label -text "Search string"
entry .search.string -width 20 -textvariable editSearchString

# create the replace string box
frame .replace
label .replace.label -text "Replace with"
entry .replace.string -width 20 -textvariable editReplaceString

# create the strip to display the search results along the scrollbar
canvas .strip -width 15 -bg black

# lay out the interface
pack .replace -side bottom -fill x
pack .replace.label -side left -padx 5 -pady 5
pack .replace.string -expand on -fill x -side left -padx 5 -pady 5

pack .search -side bottom -fill x
pack .search.label -side left -padx 5 -pady 5
pack .search.string -expand on -fill x -side left -padx 5 -pady 5

pack .txt -side left -expand true -fill both
pack .strip -side left -fill y
#pack .vsb -side left -fill y

# setup auto-search: each time the editSearchString is changed, schedule
# a call to Search and cancel any pending call
# searchString contains the current searchString (i.e., the one used be the rest of the app)
.txt tag configure found -background yellow
set searchString ""
set editSearchString ""
trace variable editSearchString w doSearch
set searchTask {}
proc doSearch {args} {
	catch {after cancel $searchTask}
	set searchTask [after 200 {
		Search $editSearchString
		set searchString $editSearchString
		if {$editReplaceString == ""} {
			set replaceString $searchString
		}
	}]
}

# setup auto-replace: each time the editReplaceString is changed, schedule
# a call to Replace and cancel any pending call
# replaceString contains the current replaceString (i.e., the one used be the rest of the app)
# in order to avoid an empty replace string, when the replace string is empty,
# is is set to the search string.
.txt tag configure replaced -background red
set replaceString ""
set editReplaceString ""
trace variable editReplaceString w doReplace
set replaceTask {}
proc doReplace {args} {
	catch {after cancel $replaceTask}
	set replaceTask [after 200 {
		if {$editReplaceString == ""} {
			Replace $searchString
			set replaceString $searchString
		} else {
			Replace $editReplaceString
			set replaceString $editReplaceString
		}
	}]
}

# each occurence is identified by a unique name
# the list of all occurence names is in Occurence
# in the text:
#	each occurence occ-i (found or replaced) has mark occ-i at its beginning
#	a found text is marked with the tag "found"
#	a replaced occurence is marked with the tag "replaced"
# in the stripbar:
#	each occurence occ-i has a tick mark with tags "tick" and "occ-i"
#	each occurence also has the tag "found" or "replaced"
#	the color of the tick is set occording to whether the occurence was found or replaced
set Occurences {}

# hilite all occurences of str in text between from and to (defaults to all text)
proc Search {str {from 1.0} {to end}} {
	global Occurences
	set occnum 0
	
	# normalize from and to
	if {$from != "1.0"} {
		set from [.txt index $from]
		set from [.txt index ${from}linestart]
	}
	if {$to != "end"} {
		set to [.txt index $to]
		set to [.txt index ${to}lineend]
	}

	# remove previous search results
	eval .txt mark unset $Occurences
	set Occurences {}
	.txt tag remove found $from end
	.txt tag remove replaced $from end
	.strip delete tick
	if {$str == ""} {
		return
	}
	
	set margin 15
	set stripscale [expr double(([winfo height .strip] - $margin*2)) / double(int([.txt index end]))]
	
	# scan text
	while {1} {
		set found [.txt search -count count -- $str $from $to]
		if {$found == ""} {
			break
		}
		
		lappend Occurences occ$occnum
		
		# set mark in text and hilite string
		.txt mark set occ$occnum $found
		.txt tag add found $found $found+${count}chars
		
		# insert tick mark in strip
		set y [expr $margin + int($found) * $stripscale]
		.strip create line 0 $y 20 $y -fill yellow -tags "tick found occ$occnum"
		
		# iterate
		set from [.txt index $found+${count}chars]
		incr occnum
	}
}

# change the replaced strings with str
proc Replace {str} {
	global Occurences
	foreach mark $Occurences {
		if {[lsearch -exact [.txt tag names $mark] "replaced"] >= 0} {
			SubstText .txt $mark replaced replaced $str red
		}
	}
}

# return the position of the occurence at or just before pos
#
proc PrevOcc {text pos} {
	set mark [$text mark previous $pos+1chars]
	if {$mark == ""} {
		return ""
	}
	while {! [string match occ* $mark]} {
		set mark [$text mark previous $mark]
		if {$mark == ""} {
			return ""
		}
	}
	return $mark
}

# return the occurence name associated to tick 'item'
# or the empty string if item is not a tick
#
proc GetOccFromItem {canvas item} {
	foreach tag [$canvas gettags $item] {
		if [string match occ* $tag] {
			return $tag
		}
	}
	return ""
}

# return the occurence name associated to a text position
# or the empty string if there is no occurence at pos
#
proc GetOccFromText {text pos} {
	# find out whether the string at pos is marked has a "found" or "replaced" tag
	foreach tag [$text tag names $pos] {
		if {$tag == "found" || $tag == "replaced"} {
			# if so return the previous mark
			return [PrevOcc $text $pos]
		}
	}
	return ""
}

# return the end indices for the found or replaced text around pos
# return {pos pos} if there is no found or replaved text around pos
# note that since mark names in the text are occurence names,
# this function can also be called with an occurence name
#
proc GetTextRange {text pos} {
	global searchString replaceString
	
	# find out whether the string at pos has a "found" or "replaced" tag
	foreach tag [$text tag names $pos] {
		if {$tag == "found"} {
			set start [PrevOcc $text $pos]
			return [list $start $start+[string length $searchString]chars]
		} elseif {$tag == "replaced"} {
			set start [PrevOcc $text $pos]
			return [list $start $start+[string length $replaceString]chars]
		}
	}
	return [list $pos $pos]
}


# scrolling
#	YScroll called by text to update scrollbar
#	YView called by scrollbar to update text

.strip lower [.strip create rectangle 2 0 17 0 -tag thumb -fill grey]

proc YScroll {args} {
	eval .vsb set $args

	foreach {from to} $args {break}
	set margin 15
	set stripscale [expr double(([winfo height .strip] - $margin*2))]
	.strip coords thumb 0 [expr $margin + $from*$stripscale] 15 [expr $margin + $to*$stripscale]
	
	# adjust items on the Mac
	global tcl_platform
	if {$tcl_platform(platform) == "macintosh"} {
		.strip move thumb 2 0
	}
}

proc YView {args} {
	eval .txt yview $args
}

# stripbar bindings
#	re-layout scrollbar when size changes
#	hilite tick mark and corresponding text when passing over it
#	goto the specified position when clicking
#	scroll the text when click-dragging
#
bind .strip <Configure> {
	Search $searchString

	# create the two buttons at top and bottom
	.strip delete button
	.strip create polygon 0 7.5 7.5 15 15 7.5 7.5 0 -fill gray -tags {button up}
	set height [winfo height .strip]
	set up [expr $height -15]
	set mid [expr $height -7.5]
	set bot $height
	.strip create polygon 0 $mid 7.5 $bot 15 $mid 7.5 $up -fill gray -tags {button down}
	
	# adjust items on the Mac
	global tcl_platform
	if {$tcl_platform(platform) == "macintosh"} {
		.strip move up 2 2
		.strip move down 2 -2
	}
	
	# update strip variables
	global Strip
	set Strip(margin) 15
	set Strip(height) [expr double(([winfo height .strip] - $Strip(margin)*2))]
	set Strip(scale) [expr $Strip(height) / double(int([.txt index end]))]
}

StateMachine::define StripBar {
	local P1 P2 YPos Delta
	local Gain InvSpeed

	state Start {
		when Enter onItem i withTag tick do {
			%W itemconfigure current -fill green
			set occ [GetOccFromItem %W current]
			eval .txt tag add current [GetTextRange .txt $occ]
		}

		when LeaveItem withTag found do {
			%W itemconfigure current -fill yellow
			eval .txt tag remove current 1.0 end
		}
		
		when LeaveItem withTag replaced do {
			%W itemconfigure current -fill red
			eval .txt tag remove current 1.0 end
		}
		
		when Down at P1 onItem i withTag up do {
			set Delta -1
			AutoScroll $Delta
			arm $InvSpeed
		} -> AutoScroll
		
		when Down at P1 onItem i withTag down do {
			set Delta 1
			AutoScroll $Delta
			arm $InvSpeed
		} -> AutoScroll
		
		when Down at P1 do {
			.txt see [Strip2TextPos [lindex $P1 1]]
			# we need to call update so the thumb is updated by the text
			# before we record its position
			update idletasks
			set YPos [lindex [%W coords thumb] 1]
		} -> Drag
	}

	state Drag {
		when Move at P2 do {
			# we need to tell the scrolled object (through YView)
			# where is the _top_ of its display.
			# therefore we need to get the current position of the top
			# of the thumb, offest it by the mouse motion, and convert it in the [0-1] range
			set dy [expr  [lindex $P2 1] - [lindex $P1 1]]
			YView moveto [Strip2UnitRange [expr $YPos + $dy]]
		}
		
		when Up -> Start
	}

	state AutoScroll {
		when TimeOut do {
			AutoScroll $Delta
			arm $InvSpeed
		}

		when Move at P2 do {
			# update Delta according to move
			set dy [expr [lindex $P2 1] - [lindex $P1 1]]
			set Delta [expr $Delta + $dy*$Gain]
			set P1 $P2
		}
		
		when Up -> Start
	}

	lproc Strip2TextPos {y} {
		global Strip
		set line [expr int(($y - $Strip(margin)) / $Strip(scale))]
		return $line.0
	}
	
	lproc Strip2UnitRange {y} {
		global Strip
		set pos [expr ($y - $Strip(margin)) / $Strip(height)]
		return $pos
	}

	lproc AutoScroll {dy} {
		# Tcl can't do int(-4.1)!!! "integer value too large to represent"
		if {$dy < 0} {
			YView scroll -[expr int(-$dy)] units
		} else {
			YView scroll [expr int($dy)] units
		}
	}
}

StateMachine::enable StripBar .strip
set StripBar::Gain 0.2
set StripBar::InvSpeed 100

# text bindings
#	hilite found string when passing over them
#

.txt tag bind found <Enter> {
	set range [GetTextRange %W current]
	eval %W tag add current $range
#	set index [lindex $range 0]
#	.strip addtag scurrent withtag $Char2Tick($index)
#	.strip itemconfig scurrent -fill green
}

.txt tag bind found <Leave> {
	eval %W tag remove current 1.0 end
#	.strip itemconfig scurrent -fill yellow
#	.strip dtag scurrent
}

.txt tag bind found <ButtonRelease-1> {
	if {$replaceString == ""} {
		SubstText %W current found replaced $searchString red
	} else {
		SubstText %W current found replaced $replaceString red
	}
}

.txt tag bind replaced <ButtonRelease-1> {
	SubstText %W current replaced found $searchString yellow
}

# substitute the text at a position pos (pos can be an index or mark)
# the found or replaced text around pos is replaced with withString,
# it tag changed from fromTag to toTag and the stripbar updated
proc SubstText {txt pos fromTag toTag withString color} {
	# find the position of the text
	set occ [GetOccFromText $txt $pos]
	set range [GetTextRange $txt $occ]
	set pos [lindex $range 0]

	# remove it, insert the new string and set the new tag
	# we need to temporarily change the gravity so the text is inserted after the occurence mark
	eval $txt delete $range
	$txt mark gravity $occ left
	$txt insert $pos $withString
	$txt mark gravity $occ right
	$txt tag add $toTag $pos $pos+[string length $withString]char

	# change the tick color
	.strip dtag $occ $fromTag
	.strip addtag $toTag withtag $occ
	.strip itemconfig $occ -fill $color
}

# make the current flag on top of the other ones
.txt tag configure current -background green
.txt tag raise current

# redefine tkTextInsert to update matched occurences
#rename tkTextInsert _tkTextInsert
proc tkTextInsert {w s} {
	if {($s == "") || ([$w cget -state] == "disabled")} {
		return
	}
	catch {
		if {[$w compare sel.first <= insert]
			&& [$w compare sel.last >= insert]} {
				$w delete sel.first sel.last
		}
	}
	$w insert insert $s
	$w see insert
	# --- trigger search
	doSearch
}

proc tkTextDelete {w} {
	if {[$w tag nextrange sel 1.0 end] != ""} {
		$w delete sel.first sel.last
	} else {
		$w delete insert-1c
		$w see insert
	}
	# --- trigger search
	doSearch
}
#console show

bind .txt <Key-BackSpace> {tkTextDelete %W; break}
bind .txt <Key-Delete> {tkTextDelete %W; break}
